import random

from c2qa.operators import CVOperators
import numpy


def allclose(a, b) -> bool:
    """Convert SciPy sparse matrices to ndarray and test with Numpy"""

    # If a and b are SciPy sparse matrices, they'll have a "toarray()" function
    if hasattr(a, "toarray"):
        a = a.toarray()

    if hasattr(b, "toarray"):
        b = b.toarray()

    return numpy.allclose(a, b)


def is_unitary_matrix(mat) -> bool:
    """Convert SciPy sparse matrix to ndarray and test with QisKit"""
    from qiskit.quantum_info.operators.predicates import is_unitary_matrix

    return is_unitary_matrix(mat.toarray())


class TestUnitary:
    """Verify operators are unitary"""

    def setup_method(self, method):
        self.cutoff = 4
        self.ops = CVOperators()

    def test_bs(self):
        assert is_unitary_matrix(self.ops.bs(random.random(), self.cutoff, self.cutoff))

    def test_d(self):
        assert is_unitary_matrix(self.ops.d(random.random(), self.cutoff))

    def test_r(self):
        assert is_unitary_matrix(self.ops.r(random.random(), self.cutoff))

    def test_s(self):
        assert is_unitary_matrix(self.ops.s(random.random(), self.cutoff))

    def test_s2(self):
        assert is_unitary_matrix(self.ops.s2(random.random(), self.cutoff, self.cutoff))


class TestMatrices:
    """Test that the operators produce the values we expect.
    TODO - would be better to test against known input & output values vs simply non-zero
    """

    def setup_method(self, method):
        self.cutoff = 4
        self.num_qumodes = 2
        self.ops = CVOperators()

    def test_a(self, capsys):
        # From https://github.com/XanaduAI/strawberryfields/blob/master/strawberryfields/backends/fockbackend/ops.py#L208-L215
        trunc = self.cutoff  # equal to CVOperators cutoff
        ret = numpy.zeros((trunc, trunc), dtype=numpy.complex128)
        for i in range(1, trunc):
            ret[i - 1][i] = numpy.sqrt(i)

        assert allclose(self.ops.get_a(self.cutoff), ret)

    def test_bs(self):
        one = self.ops.bs(1, self.cutoff, self.cutoff)
        rand = self.ops.bs(random.random(), self.cutoff, self.cutoff)

        assert not allclose(one, rand)

    def test_bs_across_os(self, capsys):
        """Doesn't actually test anything, but as it is run across platforms by GitHub
        Actions a manual comparison can be made between Linux, MacOS, and Windows"""
        with capsys.disabled():
            op = self.ops.bs(numpy.pi / 4, self.cutoff, self.cutoff)
            # print()
            # print(op)

            assert op.getnnz()

    def test_d(self, capsys):
        with capsys.disabled():
            one = self.ops.d(1, self.cutoff)
            # rand = self.ops.d(random.random())

            print()
            # print("a")
            # print(self.ops.a)
            # print("a_dag")
            # print(self.ops.a_dag)
            print("1")
            print(one)

            neg_one = self.ops.d(-1, self.cutoff)
            print("-1")
            print(neg_one)

            # assert not allclose(one, rand)

    def test_d_across_os(self, capsys):
        """Doesn't actually test anything, but as it is run across platforms by GitHub
        Actions a manual comparison can be made between Linux, MacOS, and Windows"""
        with capsys.disabled():
            op = self.ops.d(numpy.pi / 2, self.cutoff)
            # print()
            # print(op)

            assert op.nnz

    def test_r(self):
        one = self.ops.r(1, self.cutoff)
        rand = self.ops.r(random.random(), self.cutoff)

        assert not allclose(one, rand)

    def test_s(self):
        one = self.ops.s(1, self.cutoff)
        rand = self.ops.s(random.random(), self.cutoff)

        assert not allclose(one, rand)

    def test_s2(self):
        one = self.ops.s2(1, self.cutoff, self.cutoff)
        rand = self.ops.s2(random.random(), self.cutoff, self.cutoff)

        assert not allclose(one, rand)



import c2qa
import pytest
import qiskit
import math
import numpy
import random


def test_cv_c_d(capsys):
    """The cv_c_d gate should discretize all params (i.e., default behavior)"""
    with capsys.disabled():
        num_qumodes = 2
        num_qubits_per_qumode = 2
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        qr = qiskit.QuantumRegister(2)
        circuit = c2qa.CVCircuit(qmr, qr)

        theta = random.random()
        beta = -theta
        circuit.cv_c_d(theta=theta, beta=beta, qumode=qmr[0], qubit=qr[0])

        gate = circuit.data[0].operation
        total_steps = 2
        discretized_params = gate.calculate_segment_params(
            current_step=1, total_steps=total_steps, keep_state=True
        )

        print(f"Original theta={theta}")
        print(f"Discretized params {discretized_params}")

        assert discretized_params[0] == (theta / total_steps)
        assert discretized_params[1] == (beta / total_steps)


def test_cv_c_schwinger(capsys):
    """The cv_c_schwinger gate should discretize the first param, but the others not"""
    with capsys.disabled():
        num_qumodes = 2
        num_qubits_per_qumode = 2
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        qr = qiskit.QuantumRegister(2)
        circuit = c2qa.CVCircuit(qmr, qr)

        beta = random.random()
        theta_1 = random.random()
        phi_1 = random.random()
        theta_2 = random.random()
        phi_2 = random.random()
        circuit.cv_c_schwinger(
            [beta, theta_1, phi_1, theta_2, phi_2], qmr[0], qmr[1], qr[0]
        )

        gate = circuit.data[0].operation
        total_steps = 2
        discretized_params = gate.calculate_segment_params(
            current_step=1, total_steps=total_steps, keep_state=True
        )

        print(f"Original params {(beta, theta_1, phi_1, theta_2, phi_2)}")
        print(f"Discretized params {discretized_params}")

        assert discretized_params[0] == (beta / total_steps)
        assert discretized_params[1] == theta_1
        assert discretized_params[2] == phi_1
        assert discretized_params[3] == theta_2
        assert discretized_params[4] == phi_2


@pytest.mark.skip(
    reason="Enable and test manually with debug breakpoint in__calculate_segment_params to ensure only first param is discretized"
)
def test_cv_c_schwinger_animate(capsys):
    """The cv_c_schwinger gate should discretize the first param, but the others not"""
    with capsys.disabled():
        num_qumodes = 2
        num_qubits_per_qumode = 2
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        qr = qiskit.QuantumRegister(2)
        circuit = c2qa.CVCircuit(qmr, qr)

        beta = random.random()
        theta_1 = random.random()
        phi_1 = random.random()
        theta_2 = random.random()
        phi_2 = random.random()
        circuit.cv_c_schwinger(
            [beta, theta_1, phi_1, theta_2, phi_2], qmr[0], qmr[1], qr[0]
        )

        c2qa.animate.animate_wigner(
            circuit,
            file="tests/test_cv_c_schwinger_animate.gif",
            axes_min=-8,
            axes_max=8,
            animation_segments=2,
            shots=1,
        )


def test_discretize_with_pershot_statevector(capsys):
    with capsys.disabled():
        qmr = c2qa.QumodeRegister(1, 3)
        creg = qiskit.ClassicalRegister(3)
        circ = c2qa.CVCircuit(qmr, creg)
        circ.cv_initialize(7, qmr[0])

        circ.cv_delay(duration=100, qumode=qmr[0], unit="ns")
        circ.cv_measure(qmr[0], creg)

        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=0.02, circuit=circ, time_unit="ns"
        )
        state, result, fock_counts = c2qa.util.simulate(
            circ,
            noise_passes=noise_pass,
            discretize=True,
            shots=2,
            per_shot_state_vector=True,
        )

        assert result.success


def test_accumulated_counts_cv_c_r(capsys):
    def simulate_test(discretize: bool):
        qmr = c2qa.QumodeRegister(1, 3)
        anc = qiskit.circuit.AncillaRegister(1)
        cr = qiskit.circuit.ClassicalRegister(1)
        circ = c2qa.CVCircuit(qmr, anc, cr)

        circ.initialize([1, 0], anc[0])  # Ancilla in |g>
        circ.cv_initialize(3, qmr[0])  # Qumode in |3>

        # Photon number parity circuit
        circ.h(anc[0])
        circ.cv_c_r(numpy.pi / 2, qmr[0], anc[0], duration=1, unit="µs")
        circ.h(anc[0])
        circ.measure(anc[0], cr[0])

        # Simulate
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=0.1, circuit=circ, time_unit="µs"
        )

        state, result, fock_counts = c2qa.util.simulate(
            circ, noise_passes=noise_pass, discretize=discretize, shots=3000
        )
        print("##############")
        print(f"Result counts: {result.get_counts()}")
        print(f"Fock counts: {fock_counts}")
        assert result.success

    with capsys.disabled():
        print()
        print("NOT DISCRETIZED")
        simulate_test(discretize=False)
        print()
        print("DISCRETIZED")
        simulate_test(discretize=True)


def test_accumulated_counts_cv_d(capsys):
    def simulate_test(discretize: bool):
        num_qumodes = 1
        num_qubits_per_qumode = 4
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        circuit = c2qa.CVCircuit(qmr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_d(1.5, qmr[0], duration=100, unit="ns")

        photon_loss_rate = 0.02
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit=time_unit
        )

        state, result, fock_counts = c2qa.util.simulate(
            circuit, noise_passes=noise_pass, discretize=discretize, shots=200
        )
        print("##############")
        print(f"Result counts: {result.get_counts()}")
        print(f"Fock counts: {fock_counts}")
        assert result.success

    with capsys.disabled():
        print()
        print("NOT DISCRETIZED")
        simulate_test(discretize=False)
        print()
        print("DISCRETIZED")
        simulate_test(discretize=True)


def test_manual_vs_auto_discretize(capsys):
    def simulate_test(manually_discretize: bool):
        qmr = c2qa.QumodeRegister(1, 3)
        anc = qiskit.circuit.AncillaRegister(1)
        cr = qiskit.circuit.ClassicalRegister(1)
        circ = c2qa.CVCircuit(qmr, anc, cr)

        circ.initialize([1, 0], anc[0])  # Ancilla in |g>
        circ.cv_initialize(3, qmr[0])  # Qumode in |3>

        # Photon number parity circuit
        circ.h(anc[0])
        if manually_discretize:
            for _ in range(10):  # Manually discretize cv_c_r gate
                circ.cv_c_r(numpy.pi / 20, qmr[0], anc[0], duration=0.1, unit="µs")
        else:
            circ.cv_c_r(numpy.pi / 2, qmr[0], anc[0], duration=1, unit="µs")
        circ.h(anc[0])
        circ.measure(anc[0], cr[0])

        # Simulate
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=0.1, circuit=circ, time_unit="µs"
        )
        return c2qa.util.simulate(
            circ,
            noise_passes=noise_pass,
            shots=3000,
            discretize=(not manually_discretize),
        )

    with capsys.disabled():
        min_percent_diff = 99999
        max_percent_diff = 0
        for i in range(20):
            print()
            print(f"Test {i}")

            print("Manual Discretization")
            _, result_man, _ = simulate_test(manually_discretize=True)
            counts_man = result_man.get_counts()
            print(counts_man)

            print("Auto Discretization")
            _, result_auto, _ = simulate_test(manually_discretize=False)
            counts_auto = result_auto.get_counts()
            print(counts_auto)

            assert result_man.success
            assert result_auto.success

            for key in counts_man:
                max_value = max(counts_man[key], counts_auto[key])
                min_value = min(counts_man[key], counts_auto[key])
                diff = max_value - min_value
                percent_diff = diff / max_value * 100
                min_percent_diff = min(percent_diff, min_percent_diff)
                max_percent_diff = max(percent_diff, max_percent_diff)
                print(f"Key '{key}' percent difference {percent_diff}")
                assert math.isclose(counts_man[key], counts_auto[key], rel_tol=0.25)

        print(f"Min percent diff {min_percent_diff}")
        print(f"Max percent diff {max_percent_diff}")



import copy
import warnings

import numpy as np
from qiskit import QuantumCircuit, QuantumRegister
from qiskit.quantum_info.operators.predicates import is_unitary_matrix
import qiskit_aer.library.save_instructions as save

from c2qa.operators import CVOperators
from c2qa.parameterized_unitary_gate import ParameterizedUnitaryGate
from c2qa.qumoderegister import QumodeRegister


class CVCircuit(QuantumCircuit):
    """Extension of QisKit QuantumCircuit to add continuous variable (bosonic) gate support to simulations."""

    def __init__(self, *regs, name: str = None, probe_measure: bool = False):
        """Initialize the registers (at least one must be QumodeRegister) and set the circuit name.

        Args:
            name (str, optional): circuit name. Defaults to None.
            probe_measure (bool, optional): automatically support measurement with probe qubits. Defaults to False.

        Raises:
            ValueError: If no QumodeRegister is provided.
        """
        self.qmregs = []
        self._qubit_regs = []  # This needs to be unique from qregs[] in the superclass

        registers = []

        num_qubits = 0

        for reg in regs:
            if isinstance(reg, QumodeRegister):
                self.qmregs.append(reg)
                registers.append(reg.qreg)
                num_qubits += reg.size
            elif isinstance(reg, QuantumRegister):
                self._qubit_regs.append(reg)
                registers.append(reg)
                num_qubits += reg.size
            else:
                registers.append(reg)

        if len(self.qmregs) == 0:
            raise ValueError("At least one QumodeRegister must be provided.")

        # Support measurement using probe qubits
        self.probe_measure = probe_measure
        if probe_measure:
            self.probe = QuantumRegister(size=num_qubits, name="probe")
            registers.append(self.probe)

        super().__init__(*registers, name=name)

        self.ops = CVOperators()
        self.cv_snapshot_id = 0

    def merge(self, circuit: QuantumCircuit):
        """
        Merge in properties of QisKit QuantumCircuit into this instance.

        Useful if QisKit returned a new instance of QuantumCircuit after
        passing in this instance. Calling merge() can merge the two, keeping
        this instance.

        See https://qiskit.org/documentation/_modules/qiskit/circuit/quantumcircuit.html#QuantumCircuit.copy
        """

        self.qregs = circuit.qregs.copy()
        self.cregs = circuit.cregs.copy()
        self._qubits = circuit._qubits.copy()
        self._ancillas = circuit._ancillas.copy()
        self._clbits = circuit._clbits.copy()
        self._qubit_indices = circuit._qubit_indices.copy()
        self._clbit_indices = circuit._clbit_indices.copy()

        instr_instances = {id(instr): instr for instr, _, __ in circuit._data}

        instr_copies = {id_: instr.copy() for id_, instr in instr_instances.items()}

        self._data = [
            (instr_copies[id(inst)], qargs.copy(), cargs.copy())
            for inst, qargs, cargs in circuit._data
        ]

        self._calibrations = copy.deepcopy(circuit._calibrations)
        self._metadata = copy.deepcopy(circuit._metadata)

    def get_qmr_cutoff(self, qmr_index: int):
        """Return the qumode cutoff at the given index"""
        return self.qmregs[qmr_index].cutoff

    def get_qmr_num_qubits_per_qumode(self, qmr_index: int):
        """Return the number of qubits in the qumode register at the given index"""
        return self.qmregs[qmr_index].num_qubits_per_qumode

    @property
    def qumode_qubits(self):
        """All the qubits representing the qumode registers on the circuit"""
        qubits = []
        for reg in self.qmregs:
            qubits += reg[::]
        return qubits

    @property
    def qumode_qubit_indices(self):
        """A qubit index list of the qubits representing the qumode registers on the circuit"""
        qmodes = self.qumode_qubits
        indices = []

        for index, qubit in enumerate(self.qubits):
            if qubit in qmodes:
                indices.append(index)

        return indices

    @property
    def qumode_qubits_indices_grouped(self):
        """Same as qumode_qubit_indices but it groups qubits representing the same qumode together. Returns a nested list."""
        grouped_indices = []

        # Iterate through all qmregs
        for _, qmreg in enumerate(self.qmregs):
            num_qumodes_in_reg = qmreg.num_qumodes
            num_qubits_per_qumode = qmreg.num_qubits_per_qumode

            qmreg_qubit_indices = []

            # For every qubit in circuit, append index of qubit to list if qubit is in qmreg
            for qubit_index, qubit in enumerate(self.qubits):
                if qubit in qmreg[:]:
                    qmreg_qubit_indices.append(qubit_index)

            # Split list according to no. of qumodes in qmreg
            qmreg_qubit_indices = [
                qmreg_qubit_indices[
                    i * num_qubits_per_qumode : (i + 1) * num_qubits_per_qumode
                ]
                for i in range(num_qumodes_in_reg)
            ]

            # Extend final list
            grouped_indices.extend(qmreg_qubit_indices)

        return grouped_indices

    def get_qubit_index(self, qubit):
        """Return the index of the given Qubit"""
        for i, q in enumerate(self.qubits):
            if q == qubit:
                return i
        return None

    def get_qubit_indices(self, qubits: list):
        """Return the indices of the given Qubits"""
        flat_list = []
        for el in qubits:
            if isinstance(el, list):
                flat_list += el
            else:
                flat_list += [el]

        indices = []
        for i, q in enumerate(self.qubits):
            if q in flat_list:
                indices.append(i)
        return indices

    def get_qmr_index(self, qubit):
        """Return the qumode index for the given qubit. If not found, return -1."""
        for index, qmr in enumerate(self.qmregs):
            if qubit in qmr:
                return index
        raise ValueError(f"Bit {qubit} not found in circuit.")

    @property
    def cv_gate_labels(self):
        """
        All the CV gate names on the current circuit. These will be
        instances of ParameterizedUnitaryGate.
        """
        cv_gates = set()
        for instruction, qargs, cargs in self.data:
            if isinstance(instruction, ParameterizedUnitaryGate):
                cv_gates.add(instruction.label)
        return list(cv_gates)

    def cv_snapshot(self):
        """Wrap the Qiskit QuantumCircuit Snapshot function, giving it a known label for later Wigner function plot generation"""
        self.save_statevector(f"cv_snapshot_{self.cv_snapshot_id}")
        self.cv_snapshot_id += 1

    def add_qubit_register(self, *regs):
        """Add a qubit register to the circuit.

        Args:
            regs (list): List of Registers
        """
        for reg in regs:
            if isinstance(reg, QuantumRegister):
                self._qubit_regs.append(reg)
            else:
                raise ValueError(
                    "Only QuantumRegisters are allowed to be added for now"
                )
        super().add_register(*regs)

    def cv_initialize(self, params, qumodes):
        """Initialize qumode (or qumodes) to a particular state specified by params

        Args:
            params (list or int): If an int, all specified qumodes will be initialized to the Fock state with n=params.
                                  If a list, all specified qumodes will be initialized to a superposition of Fock states,
                                  with ``params[n]`` the complex amplitude of Fock state ``|n>``. The length of params must be less
                                  than or equal to the cutoff.

            qumodes (list): list of qubits representing a single qumode, or list of multiple qumodes

        Raises:
            ValueError: If the Fock state is greater than the cutoff.
        """
        # Qumodes are already represented as arrays of qubits,
        # but if this is an array of arrays, then we are initializing multiple qumodes.
        modes = qumodes
        if not isinstance(qumodes[0], list):
            modes = [qumodes]

        if isinstance(params, int):
            for qumode in modes:
                qumode_index = self.get_qmr_index(qumode[0])

                if params >= self.get_qmr_cutoff(qumode_index):
                    raise ValueError("The given Fock state is greater than the cutoff.")

                value = np.zeros(
                    (self.get_qmr_cutoff(qumode_index),), dtype=np.complex128
                )
                value[params] = 1 + 0j

                super().initialize(value, qumode)
        else:
            for qumode in modes:
                qumode_index = self.get_qmr_index(qumode[0])

                if len(params) > self.get_qmr_cutoff(qumode_index):
                    raise ValueError("len(params) exceeds the cutoff.")

                params = np.array(params) / np.linalg.norm(np.array(params))
                amplitudes = np.zeros(
                    (self.get_qmr_cutoff(qumode_index),), dtype=np.complex128
                )
                for ind in range(len(params)):
                    amplitudes[ind] = complex(params[ind])

                super().initialize(amplitudes, qumode)

    def save_circuit(self, conditional, pershot, label="statevector"):
        """Save the simulator statevector using a qiskit class"""
        return save.save_statevector(
            label=label, conditional=conditional, pershot=pershot
        )

    def cv_r(self, theta, qumode, duration=100, unit="ns"):
        """Phase space rotation gate.

        Args:
            theta (real): rotation
            qumode (list): list of qubits representing qumode
            cutoff (int): qumode cutoff

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.r,
                [theta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode),
                label="R",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode,
        )

    def cv_d(self, alpha, qumode, duration=100, unit="ns"):
        """Displacement gate.

        Args:
            alpha (real or complex): displacement
            qumode (list): list of qubits representing qumode
            cutoff (int): qumode cutoff

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.d,
                [alpha],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode),
                label="D",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode,
        )

    def cv_sq(self, theta, qumode, duration=100, unit="ns"):
        """Squeezing gate.

        Args:
            theta (real or complex): squeeze
            qumode (list): list of qubits representing qumode

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.s,
                [theta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode),
                label="S",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode,
        )

    def cv_sq2(self, theta, qumode_a, qumode_b, duration=100, unit="ns"):
        """Two-mode squeezing gate

        Args:
            theta (real or complex): squeeze
            qumode_a (list): list of qubits representing first qumode
            qumode_b (list): list of qubits representing second qumode

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.s2,
                [theta],
                cutoffs=[
                    QumodeRegister.calculate_cutoff(len(qumode_a)),
                    QumodeRegister.calculate_cutoff(len(qumode_b)),
                ],
                num_qubits=len(qumode_a) + len(qumode_b),
                label="S2",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode_a + qumode_b,
        )

    def cv_bs(self, theta, qumode_a, qumode_b, duration=100, unit="ns"):
        """Two-mode beam splitter gate.

        Args:
            theta (real or complex): beamsplitter phase
            qumode_a (list): list of qubits representing first qumode
            qumode_b (list): list of qubits representing second qumode

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.bs,
                [theta],
                cutoffs=[
                    QumodeRegister.calculate_cutoff(len(qumode_a)),
                    QumodeRegister.calculate_cutoff(len(qumode_b)),
                ],
                num_qubits=len(qumode_a) + len(qumode_b),
                label="BS",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode_a + qumode_b,
        )

    def cv_c_r(self, theta, qumode, qubit, duration=100, unit="ns"):
        """Qubit dependent phase-space rotation gate (i.e., dispersive interaction).

        Args:
            theta (real): phase
            qumode_a (list): list of qubits representing qumode
            qubit_ancilla (qubit): QisKit control qubit

        Returns:
            Instruction: QisKit instruction
        """
        self.append(
            ParameterizedUnitaryGate(
                self.ops.cr,
                [theta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="cR",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_c_rx(self, theta, qumode, qubit, duration=100, unit="ns"):
        """Qubit dependent phase-space rotation around sigma^x gate.

        Args:
            theta (real): phase
            qumode_a (list): list of qubits representing qumode
            qubit_ancilla (qubit): QisKit control qubit

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.crx,
                [theta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="cRX",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_c_ry(self, theta, qumode, qubit, duration=100, unit="ns"):
        """Qubit dependent phase-space rotation around sigma^y gate.

        Args:
            theta (real): phase
            qumode_a (list): list of qubits representing qumode
            qubit_ancilla (qubit): QisKit control qubit

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.cry,
                [theta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="cRY",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_c_d(self, theta, qumode, qubit, beta=None, duration=100, unit="ns"):
        """Conditional displacement gate.

        Args:
            theta (real): displacement
            qumode (list): list of qubits representing qumode
            qubit (Qubit): control qubit
            beta (real): By default is None, and qumode will be displaced by alpha and -alpha for qubit
            state 0 and 1, respectively. If specified, qumode will be displaced by alpha and beta for qubit state 0 and 1.

        Returns:
            Instruction: QisKit instruction
        """
        if beta is None:
            beta = -theta

        return self.append(
            ParameterizedUnitaryGate(
                self.ops.cd,
                [theta, beta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="cD",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_ecd(self, theta, qumode, qubit, duration=100, unit="ns"):
        """Echoed controlled displacement gate.

        Args:
            theta (real): displacement
            qumode (list): list of qubits representing qumode

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.ecd,
                [theta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="ECD",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_c_bs(self, theta, qumode_a, qumode_b, qubit, duration=100, unit="ns"):
        """Controlled phase two-mode beam splitter

        Args:
            theta (real or complex): phase
            qubit_ancilla (Qubit): QisKit control Qubit
            qumode_a (list): list of qubits representing first qumode
            qumode_b (list): list of qubits representing second qumode

        Returns:
            Instruction: QisKit instruction
        """
        self.append(
            ParameterizedUnitaryGate(
                self.ops.cbs,
                [theta],
                cutoffs=[
                    QumodeRegister.calculate_cutoff(len(qumode_a)),
                    QumodeRegister.calculate_cutoff(len(qumode_b)),
                ],
                num_qubits=len(qumode_a) + len(qumode_b) + 1,
                label="cBS",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode_a + qumode_b + [qubit],
        )

    def cv_c_schwinger(
        self, params, qumode_a, qumode_b, qubit, duration=100, unit="ns"
    ):
        """General form of a controlled 'Schwinger' gate, containing both the controlled phase beamsplitter
        and pairs of controlled phase space rotations as special cases.

        It has the form exp(-i*beta*(n1_hat.sigma)(n2_hat.S)),
        where ni_hat = sin(theta_i)*cos(phi_i) + sin(theta_i)*sin(phi_i) + cos(theta_i).
        sigma = [sigmax, sigmay, sigmaz] is the vector of Pauli operators, and
        S = [Sx, Sy, Sz] is a vector of Schwinger boson operators,

        Sx = (a*bdag + adag*b)/2
        Sy = (a*bdag - adag*b)/2i
        Sz = (bdag*b - adag*a)/2,

        obeying the commutation relations [Sj, Sk] = i*epsilon_{ijk}*Sz, where epsilon_{ijk} is the Levi-Civita tensor.

        Args:
            params (real): [beta, theta_1, phi_1, theta_2, phi_2]
            qubit_ancilla (Qubit): QisKit control Qubit
            qumode_a (list): list of qubits representing first qumode
            qumode_b (list): list of qubits representing second qumode

        Returns:
            Instruction: QisKit instruction
        """
        self.append(
            ParameterizedUnitaryGate(
                self.ops.cschwinger,
                params,
                cutoffs=[
                    QumodeRegister.calculate_cutoff(len(qumode_a)),
                    QumodeRegister.calculate_cutoff(len(qumode_b)),
                ],
                num_qubits=len(qumode_a) + len(qumode_b) + 1,
                label="cSchw",
                duration=duration,
                unit=unit,
                discretized_param_indices=[0],
            ),
            qargs=qumode_a + qumode_b + [qubit],
        )

    def cv_snap(self, theta, n, qumode, qubit=None, duration=100, unit="ns"):
        """SNAP (Selective Number-dependent Arbitrary Phase) gate. If no qubit is passed,
        then phases are applied to each qumode Fock state specified in theta and n (without
        explicit rotation of the qubit). If a qubit is passed, the phase will be multiplied by
        sigma_z-dependent geometric phase (akin to the implementation of the SNAP gate
        as described in Heeres et al, PRL (2015).

        Args:
            theta (real or list[real]): phase
            n (integer or list[integer]): Fock state in which the mode should acquire the phase
            qumode (list): list of qubits representing qumode
            qubit (Qubit): control qubit. If no qubit is passed, the gate will implement for sigma^z = +1.

        Returns:
            Instruction: QisKit instruction
        """
        cutoff = QumodeRegister.calculate_cutoff(len(qumode))
        if isinstance(n, int):
            if n > cutoff:
                ValueError("Fock state specified by n exceeds the cutoff.")
            if qubit is None:
                self.append(
                    ParameterizedUnitaryGate(
                        self.ops.snap,
                        [theta, n],
                        cutoffs=[cutoff],
                        num_qubits=len(qumode),
                        label="SNAP",
                        duration=duration,
                        unit=unit,
                    ),
                    qargs=qumode,
                )
            else:
                self.append(
                    ParameterizedUnitaryGate(
                        self.ops.csnap,
                        [theta, n],
                        cutoffs=[cutoff],
                        num_qubits=len(qumode) + 1,
                        label="cSNAP",
                        duration=duration,
                        unit=unit,
                    ),
                    qargs=qumode + [qubit],
                )
        elif isinstance(n, list) and isinstance(theta, list):
            if qubit is None:
                self.append(
                    ParameterizedUnitaryGate(
                        self.ops.multisnap,
                        theta + n,
                        cutoffs=[cutoff],
                        num_qubits=len(qumode),
                        label="SNAP",
                        duration=duration,
                        unit=unit,
                    ),
                    qargs=qumode,
                )
            else:
                self.append(
                    ParameterizedUnitaryGate(
                        self.ops.multicsnap,
                        theta + n,
                        cutoffs=[cutoff],
                        num_qubits=len(qumode) + 1,
                        label="cSNAP",
                        duration=duration,
                        unit=unit,
                    ),
                    qargs=qumode + [qubit],
                )
        else:
            raise ValueError(
                "if theta is passed as a list, then n must also be a list of equal length (and vice versa)."
            )

    # def cv_c_sqr(self, theta, n, qumode, qubit, duration=100, unit="ns"):
    #     """TODO"""

    # def cv_multisnap(self, thetas, ns, qumode, duration=1, unit="us"):
    #     params = thetas + ns
    #     self.append(
    #         ParameterizedUnitaryGate(
    #             self.ops.multisnap, params, num_qubits=len(qumode), label="mSNAP", duration=duration, unit=unit
    #         ),
    #         qargs=qumode,
    #     )

    def cv_c_pnr(self, max, qumode, qubit, duration=100, unit="ns"):
        """PNR (Photon number readout) TODO: Needs comments/explanation/citation!
        Args:
            max (int): the period of the mapping
            qumode (list): list of qubits representing qumode
            qubit (Qubit): control qubit.
        Returns:
            Instruction: QisKit instruction
        """
        self.append(
            ParameterizedUnitaryGate(
                self.ops.pnr,
                [max],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="c_pnr",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_eswap(self, theta, qumode_a, qumode_b, duration=100, unit="ns"):
        """Exponential SWAP gate.

        Args:
            theta (real): phase
            qumode_a (list): list of qubits representing qumode
            qumode_b (list): list of qubits representing qumode

        Returns:
            Instruction: QisKit instruction
        """
        self.append(
            ParameterizedUnitaryGate(
                self.ops.eswap,
                [theta],
                cutoffs=[
                    QumodeRegister.calculate_cutoff(len(qumode_a)),
                    QumodeRegister.calculate_cutoff(len(qumode_b)),
                ],
                num_qubits=len(qumode_a) + len(qumode_b),
                label="eSWAP",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode_a + qumode_b,
        )

    def cv_c_sq(self, theta, qumode, qubit, duration=100, unit="ns"):
        """Conditional squeezing gate.

        Args:
            theta (real or complex): squeezing ampltiude
            qumode (list): list of qubits representing qumode
            qubit (Qubit): control Qubit

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.csq,
                [theta],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="cS",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_testqubitorderf(self, phi, qubit_1, qubit_2, duration=100, unit="ns"):
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.testqubitorderf,
                [phi],
                label="testqubitorderf",
                cutoffs=[],
                num_qubits=2,
                duration=duration,
                unit=unit,
            ),
            qargs=[qubit_1] + [qubit_2],
        )

    def measure_z(self, qubit, cbit, duration=100, unit="ns"):
        """Measure qubit in z using probe qubits

        Args:
            qubit (Qubit): QisKit qubit to measure
            cbit (ClassicalBit): QisKit classical bit to measure into

        Returns:
            Instruction: QisKit measure instruction
        """
        if not self.probe_measure:
            warnings.warn(
                "Probe qubits not in use, set probe_measure to True for measure support.",
                UserWarning,
            )

        return super.measure(qubit, cbit)

    def measure_y(self, qubit, cbit):
        """Measure qubit in y using probe qubits

        Args:
            qubit (Qubit): QisKit qubit to measure
            cbit (ClassicalBit): QisKit classical bit to measure into

        Returns:
            Instruction: QisKit measure instruction
        """
        if not self.probe_measure:
            warnings.warn(
                "Probe qubits not in use, set probe_measure to True for measure support.",
                UserWarning,
            )

        self.sdg(qubit)
        self.h(qubit)
        return self.measure(qubit, cbit)

    def measure_x(self, qubit, cbit):
        """Measure qubit in x using probe qubits

        Args:
            qubit (Qubit): QisKit qubit to measure
            cbit (ClassicalBit): QisKit classical bit to measure into

        Returns:
            Instruction: QisKit measure instruction
        """
        if not self.probe_measure:
            warnings.warn(
                "Probe qubits not in use, set probe_measure to True for measure support.",
                UserWarning,
            )

        self.h(qubit)
        return self.measure(qubit, cbit)

    def cv_measure(self, qubit_qumode_list, cbit_list):
        """Measure Qumodes and Qubits in qubit_qumode_list and map onto
        classical bits specified in cbit_list.

        Args:
            qubit_qumode_list(List): List of individual Qubits and Qumodes
                (i.e., indexed elements of QubitRegisters and QumodeRegisters)
            cbit_list (List): List of classical bits to map measurements onto.
                Note: Measurement of qumodes requires log(c) classical bits,
                where c is the cutoff.
                If len(cbit_list) is greater than the required number of
                classical bits, excess will be ignored. If len(cbit_list) is
                insufficient, an error will be thrown.

        Returns:
            Instruction: QisKit measure instruction
        """

        # Flattens the list (if necessary)
        flat_list = []
        for el in qubit_qumode_list:
            if isinstance(el, list):
                flat_list += el
            else:
                flat_list += [el]

        # Check to see if too many classical registers were passed in.
        # If not, only use those needed (starting with least significant bit).
        # This piece is useful so that the user doesn't need to think about
        # how many bits are needed to read out a list of qumodes, qubits, etc.
        if len(flat_list) < len(cbit_list):
            self.measure(flat_list, cbit_list[0 : len(flat_list)])
        else:
            self.measure(flat_list, cbit_list)

    def cv_delay(self, duration, qumode, unit="ns"):
        """CV_delay. Implements an identity gate of the specified duration.
        This is particularly useful for the implementation of a noise pass.

        Args:
            duration (real): duration of delay gate
            qumode (list): list of qubits representing qumode

        Returns:
            Instruction: QisKit instruction
        """
        return self.append(
            ParameterizedUnitaryGate(
                self.ops.get_eye,
                [],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode),
                label="cv_delay(" + str(duration) + " " + unit + ")",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode,
        )

    def cv_c_multiboson_sampling(self, max, qumode, qubit, duration=1, unit="us"):
        """SNAP (Selective Number-dependent Arbitrary Phase) gates for multiboson sampling.
        Args:
            max (int): the period of the mapping
            qumode (list): list of qubits representing qumode
            qubit (Qubit): control qubit.
        Returns:
            Instruction: QisKit instruction
        """
        self.append(
            ParameterizedUnitaryGate(
                self.ops.c_multiboson_sampling,
                [max],
                cutoffs=[QumodeRegister.calculate_cutoff(len(qumode))],
                num_qubits=len(qumode) + 1,
                label="c_multiboson_sampling",
                duration=duration,
                unit=unit,
            ),
            qargs=qumode + [qubit],
        )

    def cv_gate_from_matrix(
        self,
        matrix,
        qumodes=[],
        qubits=[],
        duration=100,
        unit="ns",
        label: str = "cv_gate_from_matrix",
    ):
        """Converts matrix to gate. Note that if you choose to input some complex gate that would typically be physically
        implemented by multiple successive gate operations, PhotonLossNoisePass, simulate(discretize=True), and animate may
        not be applied in a way that is physical.

        Args:
            matrix (np.array/nested list): Matrix for conversion into gate
            qumodes (QumodeRegister/list): Qumodes initialized by QumodeRegister
            qubits (QuantumRegister/list): Qubits initialized by QuantumRegister

        Returns:
            Instruction: QisKit instruction
        """
        # If multiple qubits are given, slice to get list of qubits. Otherwise, encase QuantumRegister for single qubit in list.
        try:
            qubits = qubits[:]
        except:
            qubits = [qubits]

        # Slice QumodeRegister to get list of qumodes
        if isinstance(qumodes, QumodeRegister):
            qumodes = qumodes[:]

        # Convert matrix to np.ndarray so that we can compute error flags easier
        matrix = np.array(matrix)

        ## Error flags
        # Matrix needs to be square
        n, m = matrix.shape
        if n != m:
            raise ValueError("Matrix given is not square")

        # Determine if input matrix is same dimension as input qumodes+qubits
        if n != 2 ** (len(qumodes) + len(qubits)):
            raise ValueError("Matrix is of different dimension from qumodes + qubits")

        # Checks if input matrix is unitary
        if not is_unitary_matrix(matrix):
            raise ValueError("The mapping provided is not unitary!")

        # Make sure that np.ndarray doesn't get fed into PUG
        if isinstance(matrix, np.ndarray):
            matrix = matrix.tolist()

        # TODO Is it safe to ignore cutoff, assuming provided matrix is correct?
        cutoffs = []
        # for qumode in qumodes:
        #     cutoffs.append(QumodeRegister.calculate_cutoff(len(qumode)))

        # return self.append(
        #     ParameterizedUnitaryGate(
        #         self.ops.gate_from_matrix,
        #         [matrix],
        #         cutoffs=cutoffs,
        #         num_qubits=len(qumodes) + len(qubits),
        #         label=label,
        #         duration=duration,
        #         unit=unit
        #     ),
        #     qargs=qumodes + qubits,
        # )
        return self.unitary(matrix, qubits=qumodes + qubits, label=label)



import math
from pathlib import Path
import pytest
import random


import c2qa
import numpy as np


import qiskit
import qiskit_aer.noise as noise
from qiskit_aer.noise.noiseerror import NoiseError
from qiskit_aer.noise.passes.relaxation_noise_pass import RelaxationNoisePass
from qiskit.visualization import plot_histogram


def test_noise_model(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 2
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        qr = qiskit.QuantumRegister(2)
        cr = qiskit.ClassicalRegister(size=1)
        circuit = c2qa.CVCircuit(qmr, qr, cr)

        for qumode in range(num_qumodes):
            circuit.cv_initialize(0, qmr[qumode])

        circuit.initialize([0, 1], qr[1])  # qr[0] will init to zero

        alpha = random.random()
        circuit.cv_c_d(alpha, qmr[0], qr[0])
        circuit.cv_c_d(-alpha, qmr[0], qr[0])

        circuit.cv_c_d(-alpha, qmr[0], qr[1])
        circuit.cv_c_d(alpha, qmr[0], qr[1])

        photon_loss_rate = 1000000  # per second
        time = 5.0  # seconds
        kraus_operators = c2qa.kraus.calculate_kraus(
            photon_loss_rates=[photon_loss_rate, photon_loss_rate],
            time=time,
            circuit=circuit,
            op_qubits=[0, 1, 2],
            qumode_qubit_indices=[0, 1],
        )

        print("kraus")
        print(kraus_operators)

        state, result, fock_counts = c2qa.util.simulate(circuit)


def test_kraus_operators(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 2
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        qr = qiskit.QuantumRegister(2)
        circuit = c2qa.CVCircuit(qmr, qr)

        photon_loss_rate = 1000000  # per second
        time = 1.0  # seconds
        kraus_operators = c2qa.kraus.calculate_kraus(
            photon_loss_rates=[photon_loss_rate, photon_loss_rate],
            time=time,
            circuit=circuit,
            op_qubits=[0, 1, 2],
            qumode_qubit_indices=[0, 1],
        )

        kraus = qiskit.quantum_info.operators.channel.Kraus(kraus_operators)
        assert kraus.is_cp(), "Is not completely positive"

        print()
        print("Kraus Operators")
        accum = 0j
        for index, op in enumerate(kraus_operators):
            print(f"op {index}")
            print(op)

            op_dag = np.transpose(np.conj(op))
            print(f"op_dag {index}")
            print(op_dag)

            op_dot = np.dot(op_dag, op)
            print(f"op_dot {index}")
            print(op_dot)

            accum += op_dot
            print()

        print("Sum")
        print(accum)

        is_identity = (accum.shape[0] == accum.shape[1]) and np.allclose(
            accum, np.eye(accum.shape[0])
        )
        print(f"Sum is identity {is_identity}")
        assert is_identity, "Sum is not identity"

        assert kraus.is_tp(), "Is not trace preserving"
        assert kraus.is_cptp(), "Is not CPTP"


def test_beamsplitter_kraus_operators(capsys):
    with capsys.disabled():
        num_qumodes = 2
        qubits_per_mode = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        circuit = c2qa.CVCircuit(qmr)
        circuit.cv_initialize(2, qmr[0])
        circuit.cv_bs(1, qmr[1], qmr[0], duration=100, unit="ns")
        photon_loss_rate = 1000000  # per second
        time = 1.0  # seconds
        kraus_operators = c2qa.kraus.calculate_kraus(
            photon_loss_rates=[photon_loss_rate, photon_loss_rate],
            time=time,
            circuit=circuit,
            op_qubits=[2, 3, 0, 1],
            qumode_qubit_indices=[0, 1, 2, 3],
        )

        kraus = qiskit.quantum_info.operators.channel.Kraus(kraus_operators)
        assert kraus.is_cp(), "Is not completely positive"

        print()
        print("Kraus Operators")
        accum = 0j
        for index, op in enumerate(kraus_operators):
            print(f"op {index}")
            print(op)

            op_dag = np.transpose(np.conj(op))
            print(f"op_dag {index}")
            print(op_dag)

            op_dot = np.dot(op_dag, op)
            print(f"op_dot {index}")
            print(op_dot)

            accum += op_dot
            print()

        print("Sum")
        print(accum)

        is_identity = (accum.shape[0] == accum.shape[1]) and np.allclose(
            accum, np.eye(accum.shape[0])
        )
        print(f"Sum is identity {is_identity}")
        assert is_identity, "Sum is not identity"

        assert kraus.is_tp(), "Is not trace preserving"
        assert kraus.is_cptp(), "Is not CPTP"


def test_invalid_photon_loss_rate_length(capsys):
    with pytest.raises(Exception), capsys.disabled():
        num_qumodes = 2
        qubits_per_mode = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        init_circuit = c2qa.CVCircuit(qmr)
        init_circuit.cv_initialize(2, qmr[0])
        init_circuit.cv_bs(1, qmr[1], qmr[0], duration=100, unit="ns")
        photon_loss_rates = [1, 2, 3]  # Should only have two loss rates
        time_unit = "ns"

        # Should raise Exception for not having proper number of loss rates
        c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rates,
            circuit=init_circuit,
            time_unit=time_unit,
        )


def test_valid_photon_loss_rate_length(capsys):
    with capsys.disabled():
        num_qumodes = 2
        qubits_per_mode = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        init_circuit = c2qa.CVCircuit(qmr)
        init_circuit.cv_initialize(2, qmr[0])
        init_circuit.cv_bs(1, qmr[1], qmr[0], duration=100, unit="ns")
        photon_loss_rates = [1, 2]  # Should only have two loss rates
        time_unit = "ns"

        # Should not raise exception as has proper number of loss rates
        c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rates,
            circuit=init_circuit,
            time_unit=time_unit,
        )


def test_noise_with_beamsplitter(capsys):
    with capsys.disabled():
        num_qumodes = 2
        qubits_per_mode = 2

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        init_circuit = c2qa.CVCircuit(qmr)
        init_circuit.cv_initialize(2, qmr[0])
        init_circuit.cv_bs(1, qmr[1], qmr[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=init_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            init_circuit, noise_passes=noise_pass
        )


def test_noise_with_beamsplitter_diff_cutoff(capsys):
    with capsys.disabled():
        qmr1 = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qmr2 = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=3)
        init_circuit = c2qa.CVCircuit(qmr1, qmr2)
        init_circuit.cv_initialize(2, qmr1[0])
        init_circuit.cv_initialize(2, qmr2[0])
        init_circuit.cv_bs(1, qmr1[0], qmr2[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=init_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            init_circuit, noise_passes=noise_pass
        )


@pytest.mark.skip(reason="This test takes nearly 30 minutes to pass on Github...")
def test_noise_with_cbs_diff_cutoff(capsys):
    with capsys.disabled():
        qmr1 = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qmr2 = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=3)
        qbr = qiskit.QuantumRegister(1)
        init_circuit = c2qa.CVCircuit(qmr1, qmr2, qbr)
        init_circuit.cv_initialize(2, qmr1[0])
        init_circuit.cv_initialize(2, qmr2[0])
        init_circuit.cv_c_bs(1, qmr1[0], qmr2[0], qbr[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=init_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            init_circuit, noise_passes=noise_pass
        )


def test_noise_with_sq2_diff_cutoff(capsys):
    with capsys.disabled():
        qmr1 = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=2)
        qmr2 = c2qa.QumodeRegister(num_qumodes=1, num_qubits_per_qumode=3)
        init_circuit = c2qa.CVCircuit(qmr1, qmr2)
        init_circuit.cv_initialize(2, qmr1[0])
        init_circuit.cv_initialize(2, qmr2[0])
        init_circuit.cv_sq2(1, qmr1[0], qmr2[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=init_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            init_circuit, noise_passes=noise_pass
        )


def test_noise_with_cnd_beamsplitter(capsys):
    with capsys.disabled():
        num_qumodes = 2
        qubits_per_mode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        init_circuit = c2qa.CVCircuit(qmr, qbr)
        init_circuit.cv_initialize(2, qmr[0])
        init_circuit.cv_c_bs(1, qmr[1], qmr[0], qbr[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=init_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            init_circuit, noise_passes=noise_pass
        )


def test_photon_loss_pass_with_conditional(capsys):
    with capsys.disabled():
        num_qumodes = 1
        qubits_per_mode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        init_circuit = c2qa.CVCircuit(qmr, qbr)
        init_circuit.cv_initialize(2, qmr[0])
        init_circuit.cv_c_d(1, qmr[0], qbr[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=init_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            init_circuit, noise_passes=noise_pass
        )


def test_photon_loss_pass_delay_without_unit(capsys):
    with capsys.disabled():
        num_qumodes = 1
        qubits_per_mode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)

        fail_circuit = c2qa.CVCircuit(qmr, qbr)
        fail_circuit.cv_initialize(2, qmr[0])
        fail_circuit.delay(duration=100)  # , unit="ns")
        fail_circuit.cv_c_d(1, qmr[0], qbr[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=fail_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            fail_circuit, noise_passes=noise_pass
        )
        assert result.success


def test_photon_loss_pass_delay_with_unit(capsys):
    with capsys.disabled():
        num_qumodes = 1
        qubits_per_mode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)

        pass_circuit = c2qa.CVCircuit(qmr, qbr)
        pass_circuit.cv_initialize(2, qmr[0])
        pass_circuit.delay(duration=100, unit="ns")
        pass_circuit.cv_c_d(1, qmr[0], qbr[0], duration=100, unit="ns")
        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=pass_circuit,
            time_unit=time_unit,
        )
        state, result, fock_counts = c2qa.util.simulate(
            pass_circuit, noise_passes=noise_pass
        )
        assert result.success


def test_animate_photon_loss_pass(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 4
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        circuit = c2qa.CVCircuit(qmr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_d(0, qmr[0], duration=100, unit="ns")

        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit=time_unit
        )

        wigner_filename = "tests/test_animate_photon_loss_pass.gif"
        c2qa.animate.animate_wigner(
            circuit,
            animation_segments=10,
            file=wigner_filename,
            noise_passes=noise_pass,
        )
        assert Path(wigner_filename).is_file()


def test_animate_photon_loss_pass_with_epsilon(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 4
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        circuit = c2qa.CVCircuit(qmr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_d(0, qmr[0], duration=100, unit="ns")

        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit=time_unit
        )

        wigner_filename = "tests/test_animate_photon_loss_pas_with_epsilon.gif"
        c2qa.animate.animate_wigner(
            circuit,
            discretize_epsilon=0.1,
            file=wigner_filename,
            noise_passes=noise_pass,
        )
        assert Path(wigner_filename).is_file()


def test_photon_loss_pass_no_displacement(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 4
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        circuit = c2qa.CVCircuit(qmr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_d(0, qmr[0], duration=100, unit="ns")

        photon_loss_rate = 0.01
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit=time_unit
        )

        # state, result, fock_counts = c2qa.util.simulate(circuit, noise_passes=noise_pass)

        wigner_filename = "tests/test_photon_loss_pass_no_displacement.gif"
        c2qa.animate.animate_wigner(
            circuit,
            animation_segments=200,
            file=wigner_filename,
            noise_passes=noise_pass,
        )


def test_photon_loss_pass_slow_displacement(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 4
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        circuit = c2qa.CVCircuit(qmr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_d(1.5, qmr[0], duration=100, unit="ns")

        photon_loss_rate = 0.02
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit=time_unit
        )

        # state, result, fock_counts = c2qa.util.simulate(circuit, noise_passes=noise_pass)

        wigner_filename = "tests/test_photon_loss_pass_slow_displacement.gif"
        c2qa.animate.animate_wigner(
            circuit,
            animation_segments=200,
            file=wigner_filename,
            noise_passes=noise_pass,
            # draw_grid=True
        )


def test_photon_loss_pass_slow_conditional_displacement(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 4
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_c_d(1, qmr[0], qbr[0], duration=100, unit="ns")

        photon_loss_rate = 0.02
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit=time_unit
        )

        # state, result, fock_counts = c2qa.util.simulate(circuit, noise_passes=noise_pass)

        wigner_filename = "tests/test_photon_loss_pass_slow_conditional_displacement.gif"
        c2qa.animate.animate_wigner(
            circuit,
            animation_segments=200,
            file=wigner_filename,
            noise_passes=noise_pass,
        )


def test_photon_loss_instruction(capsys):
    with capsys.disabled():
        num_qumodes = 2
        num_qubits_per_qumode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.cv_initialize(1, qmr[0])
        circuit.cv_initialize(1, qmr[1])

        circuit.cv_d(1, qmr[0], duration=100, unit="ns")
        circuit.cv_c_d(1, qmr[1], qbr[0], duration=100, unit="ns")

        photon_loss_rate = 0.02
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=circuit,
            time_unit=time_unit,
            instructions=["cD"],
        )

        state, result, fock_counts = c2qa.util.simulate(
            circuit, noise_passes=noise_pass
        )
        assert result.success


def test_photon_loss_qumode(capsys):
    with capsys.disabled():
        num_qumodes = 2
        num_qubits_per_qumode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.cv_initialize(1, qmr[0])

        circuit.cv_d(1, qmr[0], duration=100, unit="ns")
        circuit.cv_c_d(1, qmr[1], qbr[0], duration=100, unit="ns")

        photon_loss_rate = 0.02
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=circuit,
            time_unit=time_unit,
            qumodes=qmr[1],
        )

        state, result, fock_counts = c2qa.util.simulate(
            circuit, noise_passes=noise_pass
        )
        assert result.success


def test_photon_loss_instruction_qumode(capsys):
    with capsys.disabled():
        num_qumodes = 2
        num_qubits_per_qumode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.cv_initialize(1, qmr[0])

        circuit.cv_d(1, qmr[0], duration=100, unit="ns")
        circuit.cv_c_d(1, qmr[1], qbr[0], duration=100, unit="ns")

        photon_loss_rate = 0.02
        time_unit = "ns"
        noise_pass = c2qa.kraus.PhotonLossNoisePass(
            photon_loss_rates=photon_loss_rate,
            circuit=circuit,
            time_unit=time_unit,
            instructions=["cD"],
            qumodes=qmr[0],
        )

        state, result, fock_counts = c2qa.util.simulate(
            circuit, noise_passes=noise_pass
        )
        assert result.success


def test_photon_loss_and_phase_damping(capsys):
    with capsys.disabled():
        state_a, result_a, fock_counts = _build_photon_loss_and_amp_damping_circuit(0.0)
        print(state_a)
        assert result_a.success

        state_b, result_b, fock_counts = _build_photon_loss_and_amp_damping_circuit(1.0)
        print(state_b)
        assert result_b.success

        assert not allclose(state_a, state_b)


def _build_photon_loss_and_amp_damping_circuit(amp_damp=0.3, photon_loss_rate=0.01):
    num_qumodes = 1
    qubits_per_mode = 2
    num_qubits = 1

    qmr = c2qa.QumodeRegister(
        num_qumodes=num_qumodes, num_qubits_per_qumode=qubits_per_mode
    )
    qbr = qiskit.QuantumRegister(size=num_qubits)
    circuit = c2qa.CVCircuit(qmr, qbr)
    circuit.cv_initialize(2, qmr[0])
    circuit.x(qbr[0])
    circuit.cv_d(1, qmr[0], duration=100, unit="ns")

    # Initialize phase damping NoiseModel
    noise_model = noise.NoiseModel()
    phase_error = noise.amplitude_damping_error(amp_damp)
    noise_model.add_quantum_error(phase_error, ["x"], [circuit.get_qubit_index(qbr[0])])

    # Initialize PhotonLossNoisePass
    noise_pass = c2qa.kraus.PhotonLossNoisePass(
        photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit="ns"
    )

    return c2qa.util.simulate(circuit, noise_model=noise_model, noise_passes=noise_pass)


def allclose(a, b) -> bool:
    """Convert SciPy sparse matrices to ndarray and test with Numpy"""

    # If a and b are SciPy sparse matrices, they'll have a "toarray()" function
    if hasattr(a, "toarray"):
        a = a.toarray()

    if hasattr(b, "toarray"):
        b = b.toarray()

    return np.allclose(a, b)


def test_relaxation_noise_pass(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_c_d(1, qmr[0], qbr[0], duration=100, unit="ns")

        t1s = np.ones(circuit.num_qubits).tolist()
        t2s = np.ones(circuit.num_qubits).tolist()
        noise_pass = RelaxationNoisePass(t1s, t2s)

        filename = "tests/test_relaxation_noise_pass.gif"
        c2qa.animate.animate_wigner(
            circuit,
            animation_segments=200,
            file=filename,
            noise_passes=noise_pass,
        )
        assert Path(filename).is_file()


def test_relaxation_and_photon_loss_noise_passes(capsys):
    with capsys.disabled():
        num_qumodes = 1
        num_qubits_per_qumode = 2
        num_qubits = 1

        qmr = c2qa.QumodeRegister(
            num_qumodes=num_qumodes, num_qubits_per_qumode=num_qubits_per_qumode
        )
        qbr = qiskit.QuantumRegister(size=num_qubits)
        circuit = c2qa.CVCircuit(qmr, qbr)

        circuit.cv_initialize(3, qmr[0])

        circuit.cv_c_d(1, qmr[0], qbr[0], duration=100, unit="ns")

        noise_passes = []

        t1s = np.ones(circuit.num_qubits).tolist()
        t2s = np.ones(circuit.num_qubits).tolist()
        noise_passes.append(RelaxationNoisePass(t1s, t2s))

        photon_loss_rate = 0.02
        noise_passes.append(
            c2qa.kraus.PhotonLossNoisePass(
                photon_loss_rates=photon_loss_rate, circuit=circuit, time_unit="ns"
            )
        )

        filename = "tests/test_relaxation_and_photon_loss_noise_passes.gif"
        c2qa.animate.animate_wigner(
            circuit,
            animation_segments=200,
            file=filename,
            noise_passes=noise_passes,
        )
        assert Path(filename).is_file()


def test_multi_qumode_loss_probability(capsys):
    with capsys.disabled():
        num_qumodes = 2
        num_qubits_per_qumode = 2
        qmr = c2qa.QumodeRegister(num_qumodes, num_qubits_per_qumode)
        circuit = c2qa.CVCircuit(qmr)

        circuit.cv_initialize(1, qmr[0])
        circuit.cv_initialize(1, qmr[1])
        circuit.cv_bs(np.pi / 4, qmr[0], qmr[1], duration=100, unit="ns")
        circuit.cv_bs(-np.pi / 4, qmr[0], qmr[1], duration=100, unit="ns")

        photon_loss_rate = 10000000
        noise_pass = c2qa.kraus.PhotonLossNoisePass(photon_loss_rate, circuit)

        fifty_fifty = False
        print()
        for i in range(20):
            print("----------------------")
            print(f"Iteration {i}")
            state_vector, result, fock_counts = c2qa.util.simulate(
                circuit, noise_passes=noise_pass
            )
            # plot_histogram(result.get_counts(circuit), filename=f"tests/test_manual_validate_beamsplitter-{i}.png")
            occupation, fock_states = c2qa.util.stateread(
                state_vector, 0, num_qumodes, 2**num_qubits_per_qumode, verbose=True
            )

            for qumode_state, qubit_state, amplitude in fock_states:
                # print(f"{qumode_state} {qubit_state} {amplitude}")
                qumode1 = qumode_state[0]
                qumode2 = qumode_state[1]
                probability = amplitude**2

                if (
                    (qumode1 == 1 and qumode2 == 0)
                    or (qumode1 == 0 and qumode1 == 1)
                    and math.isclose(probability, 0.5, 0.03)
                ):
                    fifty_fifty = True

        assert fifty_fifty
